from __future__ import annotations

from typing import TYPE_CHECKING

from einmesh._backends import JaxBackend
from einmesh._einmesher import _EinMesher as _BaseEinMesher
from einmesh._parser import _einmesh  # pyright: ignore[reportPrivateUsage]
from einmesh.spaces import SpaceType

if TYPE_CHECKING:
    import jax  # pyright: ignore[reportMissingImports] need for when jax is not installed


def einmesh(pattern: str, **kwargs: SpaceType) -> jax.Array:
    return _einmesh(pattern, backend=JaxBackend(), **kwargs)  # pyright: ignore[reportReturnType]


# Expose a preconfigured EinMesher class bound to JAX backend
class EinMesher(_BaseEinMesher):
    """
    EinMesher bound to the JAX backend. By default `.mesh()` will return JAX arrays
    without needing to pass a backend.
    """

    def __init__(self, pattern: str, **spaces: SpaceType) -> None:
        super().__init__(pattern, backend=JaxBackend(), **spaces)

    def sample(self) -> jax.Array | tuple[jax.Array, ...]:
        return super().sample()



from einmesh._backends import AbstractBackend
from einmesh.spaces import LinSpace
from tests.conftest import parametrize_backends


@parametrize_backends
def test_addition_with_float(backend_cls: type[AbstractBackend]):
    backend = backend_cls()

    linspace = LinSpace(start=0.0, end=1.0, num=5)

    added_linspace = linspace + 1.0
    added_linspace_reverse = 1.0 + linspace

    assert backend.allclose(linspace._sample(backend), backend.linspace(0.0, 1.0, 5))
    assert backend.allclose(added_linspace._sample(backend), backend.linspace(1.0, 2.0, 5))
    assert backend.allclose(added_linspace_reverse._sample(backend), backend.linspace(1.0, 2.0, 5))


@parametrize_backends
def test_subtraction_with_float(backend_cls: type[AbstractBackend]):
    backend = backend_cls()

    linspace = LinSpace(start=0.0, end=1.0, num=5)

    subtracted_linspace = linspace - 1.0

    assert backend.allclose(linspace._sample(backend), backend.linspace(0.0, 1.0, 5))
    assert backend.allclose(subtracted_linspace._sample(backend), backend.linspace(0.0, 1.0, 5) - 1.0)


@parametrize_backends
def test_multiplication_with_float(backend_cls: type[AbstractBackend]):
    backend = backend_cls()

    linspace = LinSpace(start=0.0, end=1.0, num=5)

    multiplied_linspace = linspace * 2.0

    assert backend.allclose(linspace._sample(backend), backend.linspace(0.0, 1.0, 5))
    assert backend.allclose(multiplied_linspace._sample(backend), backend.linspace(0.0, 2.0, 5))


@parametrize_backends
def test_truediv_with_float(backend_cls: type[AbstractBackend]):
    backend = backend_cls()

    linspace = LinSpace(start=0.0, end=1.0, num=5)

    divided_linspace = linspace / 2.0

    assert backend.allclose(linspace._sample(backend), backend.linspace(0.0, 1.0, 5))
    assert backend.allclose(divided_linspace._sample(backend), backend.linspace(0.0, 0.5, 5))


@parametrize_backends
def test_mod_with_float(backend_cls: type[AbstractBackend]):
    backend = backend_cls()

    linspace = LinSpace(start=0.0, end=1.0, num=5)

    mod_linspace = linspace % 2.0

    assert backend.allclose(linspace._sample(backend), backend.linspace(0.0, 1.0, 5))
    assert backend.allclose(mod_linspace._sample(backend), backend.linspace(0.0, 1.0, 5) % 2.0)


@parametrize_backends
def test_floordiv_with_float(backend_cls: type[AbstractBackend]):
    backend = backend_cls()

    linspace = LinSpace(start=0.0, end=1.0, num=5)

    floordiv_linspace = linspace // 2.0

    assert backend.allclose(linspace._sample(backend), backend.linspace(0.0, 1.0, 5))
    assert backend.allclose(floordiv_linspace._sample(backend), backend.linspace(0.0, 1.0, 5) // 2.0)


@parametrize_backends
def test_pow_with_float(backend_cls: type[AbstractBackend]):
    backend = backend_cls()

    linspace = LinSpace(start=0.0, end=1.0, num=5)

    pow_linspace = linspace**2.0

    assert backend.allclose(linspace._sample(backend), backend.linspace(0.0, 1.0, 5))
    assert backend.allclose(pow_linspace._sample(backend), backend.linspace(0.0, 1.0, 5) ** 2.0)


@parametrize_backends
def test_negation(backend_cls: type[AbstractBackend]):
    backend = backend_cls()

    linspace = LinSpace(start=0.0, end=1.0, num=5)

    negated_linspace = -linspace

    assert backend.allclose(negated_linspace._sample(backend), backend.linspace(0.0, -1.0, 5))


@parametrize_backends
def test_pos(backend_cls: type[AbstractBackend]):
    backend = backend_cls()

    linspace = LinSpace(start=0.0, end=1.0, num=5)

    pos_linspace = +linspace

    assert backend.allclose(pos_linspace._sample(backend), backend.linspace(0.0, 1.0, 5))



from __future__ import annotations

from typing import TYPE_CHECKING

from einmesh._backends import TorchBackend
from einmesh._einmesher import _EinMesher
from einmesh._parser import _einmesh  # pyright: ignore[reportPrivateUsage]
from einmesh.spaces import SpaceType

if TYPE_CHECKING:
    import torch  # pyright: ignore[reportMissingImports]


def einmesh(pattern: str, **kwargs: SpaceType) -> torch.Tensor:
    return _einmesh(pattern, backend=TorchBackend(), **kwargs)  # pyright: ignore[reportReturnType]


# Expose a preconfigured EinMesher class bound to Torch backend
class EinMesher(_EinMesher):
    """
    EinMesher bound to the Torch backend. By default `.mesh()` will return PyTorch tensors
    without needing to pass a backend.
    """

    def __init__(self, pattern: str, **spaces: SpaceType):
        super().__init__(pattern, backend=TorchBackend(), **spaces)

    def sample(self) -> torch.Tensor | tuple[torch.Tensor, ...]:
        return super().sample()



import pytest

from einmesh._backends import JaxBackend, NumpyBackend, TorchBackend

# Define the reusable decorator
parametrize_backends = pytest.mark.parametrize(
    argnames="backend_cls",
    argvalues=[
        pytest.param(
            TorchBackend,
            marks=pytest.mark.skipif(condition=not TorchBackend.is_available(), reason="Torch not available"),
        ),
        pytest.param(
            NumpyBackend,
            marks=pytest.mark.skipif(condition=not NumpyBackend.is_available(), reason="Numpy not available"),
        ),
        pytest.param(
            JaxBackend, marks=pytest.mark.skipif(condition=not JaxBackend.is_available(), reason="Jax not available")
        ),
    ],
    ids=["torch", "numpy", "jax"],  # Optional: Add IDs for clearer test names
)



from typing import Any

import einops
import pytest

from einmesh._backends import AbstractBackend
from einmesh._parser import UndefinedSpaceError
from einmesh._parser import _einmesh as einmesh
from einmesh.spaces import (
    ConstantSpace,
    LinSpace,
    ListSpace,
    LogSpace,
    NormalDistribution,
    RangeSpace,
    UniformDistribution,
)

# Try explicit absolute import
from tests.conftest import parametrize_backends


@parametrize_backends
def test_linear_space(backend_cls: type[AbstractBackend]):
    # Test initialization
    backend = backend_cls()
    lin_space = LinSpace(start=0.0, end=1.0, num=5)
    assert lin_space.start == 0.0
    assert lin_space.end == 1.0
    assert lin_space.num == 5

    # Test sampling
    samples = lin_space._sample(backend)
    assert backend.is_appropriate_type(samples)
    assert samples.shape == (5,)
    assert backend.allclose(samples, backend.linspace(0.0, 1.0, 5))


@parametrize_backends
def test_log_space(backend_cls: type[AbstractBackend]):
    # Test initialization
    backend = backend_cls()
    log_space = LogSpace(start=0.0, end=1.0, num=5, base=10)
    assert log_space.start == 0.0
    assert log_space.end == 1.0
    assert log_space.num == 5
    assert log_space.base == 10

    # Test sampling
    samples = log_space._sample(backend)
    assert backend.is_appropriate_type(samples)
    assert samples.shape == (5,)
    assert backend.allclose(samples, backend.logspace(0.0, 1.0, 5, base=10))


@parametrize_backends
def test_normal_distribution(backend_cls: type[AbstractBackend]):
    # Test initialization
    backend = backend_cls()
    normal_dist = NormalDistribution(mean=0.0, std=1.0, num=1000)
    assert normal_dist.mean == 0.0
    assert normal_dist.std == 1.0
    assert normal_dist.num == 1000

    # Test sampling
    samples = normal_dist._sample(backend)
    assert backend.is_appropriate_type(samples)
    assert samples.shape == (1000,)

    # Statistical tests (approximate due to random nature)
    assert abs(samples.mean().item() - normal_dist.mean) < 0.1
    assert abs(samples.std().item() - normal_dist.std) < 0.1


@parametrize_backends
def test_uniform_distribution(backend_cls: type[AbstractBackend]):
    # Test initialization
    backend = backend_cls()
    uniform_dist = UniformDistribution(low=-1.0, high=1.0, num=1000)
    assert uniform_dist.low == -1.0
    assert uniform_dist.high == 1.0
    assert uniform_dist.num == 1000

    # Test sampling
    samples = uniform_dist._sample(backend)
    assert backend.is_appropriate_type(samples)
    assert backend.shape(samples) == (1000,)

    # Check bounds - adapt comparison for backend
    assert backend.all(samples >= -1.0)
    assert backend.all(samples <= 1.0)


@parametrize_backends
def test_constant_space(backend_cls: type[AbstractBackend]):
    # Test initialization
    backend = backend_cls()
    const_space = ConstantSpace(value=5.0, num=3)
    assert const_space.value == 5.0
    assert const_space.num == 3

    # Test sampling
    samples = const_space._sample(backend)
    assert backend.is_appropriate_type(samples)
    assert backend.shape(samples) == (3,)
    assert backend.allclose(samples, backend.tensor([5.0, 5.0, 5.0]))

    # Test default num=1
    const_space_single = ConstantSpace(value=-2.0)
    assert const_space_single.num == 1
    samples_single = const_space_single._sample(backend)
    assert samples_single.shape == (1,)
    assert backend.allclose(samples_single, backend.tensor([-2.0]))


@parametrize_backends
def test_list_space(backend_cls: type[AbstractBackend]):
    # Test initialization
    backend = backend_cls()
    test_values = [1.1, 2.2, 3.3, 4.4]
    list_space = ListSpace(values=test_values)
    assert list_space.values == test_values

    # Test sampling
    samples = list_space._sample(backend)
    assert backend.is_appropriate_type(samples)
    assert samples.shape == (len(test_values),)

    # Convert test values to tensor using appropriate backend
    expected = backend.tensor(test_values)
    assert backend.allclose(samples, expected)


@parametrize_backends
@pytest.mark.parametrize(
    "start, stop, step, expected",
    [
        (0, 5, 1, [0, 1, 2, 3, 4]),
        (1, 6, 1, [1, 2, 3, 4, 5]),
        (-5, 0, 1, [-5, -4, -3, -2, -1]),
        (0, 10, 2, [0, 2, 4, 6, 8]),
        (10, 0, -2, [10, 8, 6, 4, 2]),
        (-6, 6, 3, [-6, -3, 0, 3]),
        (5, -5, -2, [5, 3, 1, -1, -3]),
        (0, 1, 0.25, [0, 0.25, 0.5, 0.75]),
    ],
)
def test_range_space(backend_cls: type[AbstractBackend], start, stop, step, expected):
    # Test initialization
    backend = backend_cls()
    range_space = RangeSpace(start=start, stop=stop, step=step)
    assert range_space.start == start
    assert range_space.stop == stop
    assert range_space.num is not None

    # Test basic arange functionality
    samples = range_space._sample(backend)
    assert backend.allclose(samples, backend.arange(start, stop, step))


@parametrize_backends
def test_einmesh_integration(backend_cls: type[AbstractBackend]):
    # Test einmesh with multiple spaces
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LogSpace(0.0, 1.0, 3)

    meshes = einmesh("x y", x=x_space, y=y_space, backend=backend)

    assert len(meshes) == 2
    assert backend.is_appropriate_type(meshes[0])
    assert backend.is_appropriate_type(meshes[1])
    assert meshes[0].shape == (5, 3)
    assert meshes[1].shape == (5, 3)


@parametrize_backends
def test_linsspace_integration(backend_cls: type[AbstractBackend]):
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)

    meshes = einmesh("x y", x=x_space, y=y_space, backend=backend)

    assert len(meshes) == 2


@parametrize_backends
def test_logspace_integration(backend_cls: type[AbstractBackend]):
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LogSpace(0.0, 1.0, 3)

    meshes = einmesh("x y", x=x_space, y=y_space, backend=backend)

    assert len(meshes) == 2


@parametrize_backends
def test_normaldistribution_integration(backend_cls: type[AbstractBackend]):
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = NormalDistribution(0.0, 1.0, 3)

    meshes = einmesh("x y", x=x_space, y=y_space, backend=backend)

    assert len(meshes) == 2


@parametrize_backends
def test_uniformdistribution_integration(backend_cls: type[AbstractBackend]):
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = UniformDistribution(0.0, 1.0, 3)

    meshes = einmesh("x y", x=x_space, y=y_space, backend=backend)

    assert len(meshes) == 2


@parametrize_backends
def test_constantspace_integration(backend_cls: type[AbstractBackend]):
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = ConstantSpace(value=7.0, num=3)

    meshes = einmesh("x y", x=x_space, y=y_space, backend=backend)

    assert len(meshes) == 2
    assert meshes[0].shape == (5, 3)
    assert meshes[1].shape == (5, 3)
    # Check if y values are constant
    assert backend.all(meshes[1] == 7.0)


@parametrize_backends
def test_listspace_integration(backend_cls: type[AbstractBackend]) -> None:
    backend = backend_cls()
    x_space: LinSpace = LinSpace(start=0.0, end=1.0, num=5)
    list_values: list[float] = [10.0, 20.0, 30.0]
    y_space: ListSpace = ListSpace(values=list_values)

    meshes: Any | tuple[Any, ...] = einmesh("x y", x=x_space, y=y_space, backend=backend)

    assert len(meshes) == 2
    assert meshes[0].shape == (5, 3)
    assert meshes[1].shape == (5, 3)
    # Check if y values match the list across the x dimension
    expected_y = einops.repeat(backend.tensor(list_values), "n -> h n", h=5)
    assert backend.allclose(meshes[1], expected_y)


@parametrize_backends
def test_invalid_space(backend_cls: type[AbstractBackend]):
    backend = backend_cls()
    with pytest.raises(UndefinedSpaceError) as exc_info:
        einmesh("x y", x=LinSpace(0.0, 1.0, 5), backend=backend)  # Missing y space
    assert str(exc_info.value) == "Undefined space: y"



from __future__ import annotations

import importlib.util
import sys
from abc import ABC, abstractmethod
from typing import Any, Literal

from einmesh._exceptions import UnknownBackendError
from einmesh.types import Number

_loaded_backends: dict[str, AbstractBackend] = {}
_type2backend: dict[type, AbstractBackend] = {}


def get_backend(tensor) -> AbstractBackend:
    """
    Takes a correct backend (e.g. numpy backend if tensor is numpy.ndarray) for a tensor.
    If needed, imports package and creates backend
    """
    _type = type(tensor)
    _result = _type2backend.get(_type)
    if _result is not None:
        return _result

    for _framework_name, backend in list(_loaded_backends.items()):
        if backend.is_appropriate_type(tensor):
            _type2backend[_type] = backend
            return backend

    # Find backend subclasses recursively
    backend_subclasses = []
    backends = AbstractBackend.__subclasses__()
    while backends:
        backend = backends.pop()
        backends += backend.__subclasses__()
        backend_subclasses.append(backend)

    for BackendSubclass in backend_subclasses:
        if BackendSubclass.framework_name not in _loaded_backends and BackendSubclass.framework_name in sys.modules:
            # check that module was already imported. Otherwise it can't be imported
            backend = BackendSubclass()
            _loaded_backends[backend.framework_name] = backend
            if backend.is_appropriate_type(tensor):
                _type2backend[_type] = backend
                return backend

    raise UnknownBackendError(backend=type(tensor).__name__)


class AbstractBackend(ABC):
    """Base backend class, major part of methods are only for debugging purposes."""

    framework_name: str

    def is_appropriate_type(self, tensor):
        """helper method should recognize tensors it can handle"""
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def is_available() -> bool:
        raise NotImplementedError()

    def __repr__(self):
        return f"<einmesh backend for {self.framework_name}>"

    @abstractmethod
    def linspace(self, start: float, stop: float, num: int):
        raise NotImplementedError()

    @abstractmethod
    def logspace(self, start: float, stop: float, num: int, base: float = 10):
        raise NotImplementedError()

    @abstractmethod
    def arange(self, start: Number, stop: Number, step: Number):
        raise NotImplementedError()

    @abstractmethod
    def full(self, shape, fill_value):
        raise NotImplementedError()

    @abstractmethod
    def tensor(self, data, dtype=None):
        raise NotImplementedError()

    @abstractmethod
    def normal(self, mean, std, size):
        raise NotImplementedError()

    @abstractmethod
    def rand(self, size) -> Any:
        raise NotImplementedError()

    @abstractmethod
    def concat(self, tensors, axis: int):
        raise NotImplementedError()

    @abstractmethod
    def allclose(self, a, b, rtol=1e-05, atol=1e-08, equal_nan=False):
        raise NotImplementedError()

    @abstractmethod
    def shape(self, tensor) -> tuple[int, ...]:
        raise NotImplementedError()

    @abstractmethod
    def meshgrid(self, *tensors, indexing: Literal["ij", "xy"] = "ij"):
        raise NotImplementedError()

    @abstractmethod
    def size(self, tensor):
        raise NotImplementedError()

    @abstractmethod
    def all(self, tensor):
        raise NotImplementedError()

    @abstractmethod
    def stack(self, tensors, dim: int):
        raise NotImplementedError()

    @property
    @abstractmethod
    def float(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def float32(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def float64(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def int8(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def int16(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def int32(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def int64(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def bool(self):
        raise NotImplementedError()

    @abstractmethod
    def abs(self, x):
        raise NotImplementedError()

    @abstractmethod
    def add(self, x, summand):
        raise NotImplementedError()

    @abstractmethod
    def sub(self, x, subtrahend):
        raise NotImplementedError()

    @abstractmethod
    def mul(self, x, multiplier):
        raise NotImplementedError()

    @abstractmethod
    def truediv(self, x, divisor):
        raise NotImplementedError()

    @abstractmethod
    def mod(self, x, divisor):
        raise NotImplementedError()

    @abstractmethod
    def floordiv(self, x, divisor):
        raise NotImplementedError()

    @abstractmethod
    def pos(self, x):
        raise NotImplementedError()

    @abstractmethod
    def neg(self, x):
        raise NotImplementedError()

    @abstractmethod
    def pow(self, base, exponent):
        raise NotImplementedError()

    @abstractmethod
    def sin(self, x):
        raise NotImplementedError()

    @abstractmethod
    def cos(self, x):
        raise NotImplementedError()

    @abstractmethod
    def exp(self, x):
        raise NotImplementedError()

    @abstractmethod
    def any(self, x):
        raise NotImplementedError()


class NumpyBackend(AbstractBackend):
    framework_name = "numpy"

    def __init__(self):
        import numpy  # pyright: ignore[reportMissingImports]

        self.np = numpy

    @staticmethod
    def is_available() -> bool:
        return importlib.util.find_spec(name="numpy") is not None

    def is_appropriate_type(self, tensor):
        return isinstance(tensor, self.np.ndarray)

    def linspace(self, start, stop, num):
        return self.np.linspace(start, stop, num)

    def concat(self, tensors, axis: int):
        return self.np.concatenate(tensors, axis=axis)

    def logspace(self, start, stop, num, base=10):
        return self.np.logspace(start, stop, num, base=base)

    def full(self, shape, fill_value):
        return self.np.full(shape, fill_value)

    def tensor(self, data, dtype=None):
        return self.np.array(data, dtype=dtype)

    def normal(self, mean, std, size):
        return self.np.random.normal(mean, std, size=size)

    def rand(self, size):
        if isinstance(size, int):
            size = (size,)
        return self.np.random.rand(*size)

    def arange(self, start: float | int | None, stop: float | int, step: float | int | None = None):
        return self.np.arange(start, stop, step)

    def allclose(self, a, b, rtol=1e-05, atol=1e-08, equal_nan=False):
        return self.np.allclose(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan)

    def shape(self, tensor):
        return tuple(tensor.shape)

    def meshgrid(self, *tensors, indexing: Literal["ij", "xy"] = "ij"):
        return self.np.meshgrid(*tensors, indexing=indexing)

    def size(self, tensor):
        return tensor.size

    def all(self, tensor):
        return tensor.all()

    def stack(self, tensors, dim: int):
        return self.np.stack(tensors, axis=dim)

    @property
    def float(self):
        return self.np.float64

    @property
    def float32(self):
        return self.np.float32

    @property
    def float64(self):
        return self.np.float64

    @property
    def int8(self):
        return self.np.int8

    @property
    def int16(self):
        return self.np.int16

    @property
    def int32(self):
        return self.np.int32

    @property
    def int64(self):
        return self.np.int64

    @property
    def bool(self):
        return self.np.bool_

    def abs(self, x):
        return self.np.abs(x)

    def add(self, x, summand):
        return self.np.add(x, summand)

    def sub(self, x, subtrahend):
        return self.np.subtract(x, subtrahend)

    def mul(self, x, multiplier):
        return self.np.multiply(x, multiplier)

    def truediv(self, x, divisor):
        return self.np.true_divide(x, divisor)

    def mod(self, x, divisor):
        return self.np.mod(x, divisor)

    def floordiv(self, x, divisor):
        return self.np.floor_divide(x, divisor)

    def pow(self, base, exponent):
        return self.np.power(base, exponent)

    def pos(self, x):
        return self.np.positive(x)

    def neg(self, x):
        return self.np.negative(x)

    def sin(self, x):
        return self.np.sin(x)

    def cos(self, x):
        return self.np.cos(x)

    def exp(self, x):
        return self.np.exp(x)

    def any(self, x):
        return self.np.any(x)


class JaxBackend(NumpyBackend):
    framework_name = "jax"

    def __init__(self):
        super().__init__()
        # Preserve original numpy module for seed generation
        self.onp = self.np

        import jax.numpy  # pyright: ignore[reportMissingImports]
        import jax.random  # pyright: ignore[reportMissingImports]

        self.np = jax.numpy
        self._random = jax.random
        # Initialize PRNGKey with random seed to ensure varied sampling across instances
        import random

        seed = random.randrange(0, 2**31 - 1)
        self.key = self._random.PRNGKey(seed)

    @staticmethod
    def is_available() -> bool:
        return importlib.util.find_spec(name="jax") is not None

    def rand(self, size) -> Any:
        new_key, subkey = self._random.split(self.key)
        self.key = new_key
        return self._random.uniform(subkey, size)

    def normal(self, mean, std, size):
        new_key, subkey = self._random.split(self.key)
        self.key = new_key
        return self._random.normal(key=subkey, shape=size) * std + mean


class TorchBackend(AbstractBackend):
    framework_name = "torch"

    def __init__(self):
        import torch  # pyright: ignore[reportMissingImports]

        self.torch = torch

    @staticmethod
    def is_available() -> bool:
        return importlib.util.find_spec(name="torch") is not None

    def is_appropriate_type(self, tensor):
        return isinstance(tensor, self.torch.Tensor)

    def arange(self, start: Number, stop: Number, step: Number):
        return self.torch.arange(start=start, end=stop, step=step)

    def linspace(self, start, stop, num, dtype=None):
        return self.torch.linspace(start, stop, num, dtype=dtype)

    def logspace(self, start, stop, num, base=10):
        return self.torch.logspace(start, stop, num, base=base)

    def full(self, shape, fill_value):
        return self.torch.full(shape, fill_value)

    def tensor(self, data, dtype=None):
        return self.torch.tensor(data, dtype=dtype)

    def normal(self, mean, std, size):
        return self.torch.normal(mean, std, size=size)

    def rand(self, size):
        return self.torch.rand(size)

    def concat(self, tensors, axis: int):
        return self.torch.cat(tensors, dim=axis)

    def allclose(self, a, b, rtol=1e-05, atol=1e-08, equal_nan=False):
        return self.torch.allclose(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan)

    def shape(self, tensor):
        return tuple(tensor.shape)

    def meshgrid(self, *tensors, indexing="ij"):
        # Ensure all tensors have the same dtype (float32) to avoid meshgrid dtype mismatch
        dt = self.float32
        tensors_cast = [t.to(dt) for t in tensors]
        return self.torch.meshgrid(*tensors_cast, indexing=indexing)

    def size(self, tensor):
        return tensor.numel()

    def all(self, tensor):
        return tensor.all()

    def stack(self, tensors, dim: int):
        return self.torch.stack(tensors, dim=dim)

    @property
    def float(self):
        return self.torch.float

    @property
    def float32(self):
        return self.torch.float32

    @property
    def float64(self):
        return self.torch.float64

    @property
    def int(self):
        return self.torch.int

    @property
    def int8(self):
        return self.torch.int8

    @property
    def int16(self):
        return self.torch.int16

    @property
    def int32(self):
        return self.torch.int32

    @property
    def int64(self):
        return self.torch.int64

    @property
    def bool(self):
        return self.torch.bool

    def abs(self, x):
        return self.torch.abs(x)

    def add(self, x, summand):
        return self.torch.add(x, summand)

    def sub(self, x, subtrahend):
        return self.torch.sub(x, subtrahend)

    def mul(self, x, multiplier):
        return self.torch.mul(x, multiplier)

    def truediv(self, x, divisor):
        return self.torch.true_divide(x, divisor)

    def mod(self, x, divisor):
        return self.torch.fmod(x, divisor)

    def floordiv(self, x, divisor):
        return self.torch.floor_divide(x, divisor)

    def pow(self, base, exponent):
        return self.torch.pow(base, exponent)

    def pos(self, x):
        return self.torch.positive(x)

    def neg(self, x):
        return self.torch.negative(x)

    def sin(self, x):
        return self.torch.sin(x)

    def cos(self, x):
        return self.torch.cos(x)

    def exp(self, x):
        return self.torch.exp(x)

    def any(self, x):
        return self.torch.any(x)



from einmesh._backends import AbstractBackend
from einmesh._parser import _einmesh as einmesh
from einmesh.spaces import LinSpace
from tests.conftest import parametrize_backends


@parametrize_backends
def test_linear_space(backend_cls: type[AbstractBackend]):
    backend = backend_cls()
    x_torch = backend.linspace(0, 1, 10)

    x_einmesh = einmesh("i", i=LinSpace(start=0, end=1, num=10), backend=backend)

    assert backend.allclose(x_torch, x_einmesh[0])


@parametrize_backends
def test_linear_space_2d(backend_cls: type[AbstractBackend]):
    backend = backend_cls()
    x_torch = backend.linspace(0, 1, 10)
    y_torch = backend.linspace(0, 1, 10)

    x_torch, y_torch = backend.meshgrid(x_torch, y_torch, indexing="ij")

    x_einmesh = einmesh("i j", i=LinSpace(start=0, end=1, num=10), j=LinSpace(start=0, end=1, num=10), backend=backend)

    assert backend.allclose(x_torch, x_einmesh[0])
    assert backend.allclose(y_torch, x_einmesh[1])


@parametrize_backends
def test_linear_space_8d(backend_cls: type[AbstractBackend]):
    backend = backend_cls()
    dims = [backend.linspace(0, 1, 10) for _ in range(8)]
    torch_meshes = backend.meshgrid(*dims, indexing="ij")

    einmesh_spaces = {f"dim{i}": LinSpace(start=0, end=1, num=10) for i in range(8)}
    einmesh_pattern = " ".join(einmesh_spaces.keys())
    einmesh_meshes = einmesh(einmesh_pattern, backend=backend, **einmesh_spaces)

    for torch_mesh, einmesh_mesh in zip(torch_meshes, einmesh_meshes):
        assert backend.allclose(torch_mesh, einmesh_mesh)



from __future__ import annotations

import math
from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any

from einmesh._backends import AbstractBackend
from einmesh._operators import BackendOperator, OperatorFactory


@dataclass
class SpaceType(ABC):
    """Base class for all space types."""

    operators: list[BackendOperator] = field(init=False, default_factory=list)

    def _with_operator(self, operator: BackendOperator, prepend: bool = False) -> SpaceType:
        """Return a copy of this space with the given operator applied."""
        new_space = deepcopy(self)
        if prepend:
            new_space.operators.insert(0, operator)
        else:
            new_space.operators.append(operator)
        return new_space

    def _with_operators(self, operators: list[BackendOperator], prepend: bool = False) -> SpaceType:
        """Return a copy of this space with multiple operators applied."""
        new_space = deepcopy(self)
        if prepend:
            new_space.operators = operators + new_space.operators
        else:
            new_space.operators.extend(operators)
        return new_space

    def __abs__(self) -> SpaceType:
        return self._with_operator(OperatorFactory.abs())

    def __add__(self, other) -> SpaceType:
        return self._with_operator(OperatorFactory.add(value=other))

    def __radd__(self, other) -> SpaceType:
        return self.__add__(other)

    def __sub__(self, other) -> SpaceType:
        return self._with_operator(OperatorFactory.sub(value=other))

    def __rsub__(self, other) -> SpaceType:
        return self._with_operators(
            operators=[
                OperatorFactory.neg(),
                OperatorFactory.add(value=other),
            ],
            prepend=True,
        )

    def __mul__(self, other) -> SpaceType:
        return self._with_operator(OperatorFactory.mul(value=other))

    def __rmul__(self, other) -> SpaceType:
        return self.__mul__(other)

    def __mod__(self, other) -> SpaceType:
        return self._with_operator(OperatorFactory.mod(value=other))

    def __floordiv__(self, other) -> SpaceType:
        return self._with_operator(OperatorFactory.floor_div(value=other))

    def __truediv__(self, other) -> SpaceType:
        return self._with_operator(OperatorFactory.div(value=other))

    def __neg__(self) -> SpaceType:
        return self._with_operator(OperatorFactory.neg())

    def __pos__(self) -> SpaceType:
        return self._with_operator(OperatorFactory.pos())

    def __pow__(self, other) -> SpaceType:
        return self._with_operator(OperatorFactory.pow(exponent=other))

    @abstractmethod
    def _generate_samples(self, backend: AbstractBackend):
        """Create a new space type."""
        ...

    def _sample(self, backend: AbstractBackend) -> Any:
        """Sample the space type."""
        sample = self._generate_samples(backend)
        return self._apply_operators(sample, backend)

    def _apply_operators(self, sample, backend: AbstractBackend) -> Any:
        """Apply the operators to the space type."""
        for operator in self.operators:
            sample = operator._apply(x=sample, backend=backend)
        return sample


@dataclass
class LogSpace(SpaceType):
    """
    Represents a sequence of points spaced logarithmically.

    This class generates a tensor of `num` points between `10**start` and `10**end`
    (or `base**start` and `base**end` if `base` is specified), spaced
    logarithmically.

    Attributes:
        start: The starting exponent of the sequence.
        end: The ending exponent of the sequence.
        num: The number of points to generate.
        base: The base of the logarithm. Defaults to 10.
    """

    start: float
    end: float
    num: int
    base: float = 10

    def _generate_samples(self, backend: AbstractBackend):
        """Generates the logarithmically spaced points."""
        sample = backend.logspace(self.start, self.end, self.num, base=self.base)
        return self._apply_operators(sample, backend)


@dataclass
class LgSpace(LogSpace):
    """
    Represents a sequence of points spaced in base 2.
    """

    start: float
    end: float
    num: int
    base: float = field(init=False, default=2)

    def __post_init__(self) -> None:
        object.__setattr__(self, "base", 2)


@dataclass
class LnSpace(LogSpace):
    """
    Represents a sequence of points spaced in base e.
    """

    start: float
    end: float
    num: int
    base: float = field(init=False, default=math.e)

    def __post_init__(self) -> None:
        object.__setattr__(self, "base", math.e)


@dataclass
class LinSpace(SpaceType):
    """
    Represents a sequence of points spaced linearly.

    This class generates a tensor of `num` points evenly spaced between `start`
    and `end` (inclusive).

    Attributes:
        start: The starting value of the sequence.
        end: The ending value of the sequence.
        num: The number of points to generate.
    """

    start: float
    end: float
    num: int

    def _generate_samples(self, backend: AbstractBackend):
        """Generates the linearly spaced points."""
        return backend.linspace(self.start, self.end, self.num)


@dataclass
class RangeSpace(SpaceType):
    """
    Return evenly spaced values within a given interval.

    This class generates a tensor of `num` points evenly spaced between `start`
    and `end` (inclusive). It works like a arange function of numpy, torch or jax.

    Attributes:
        start: The starting value of the sequence.
        end: The ending value of the sequence.
        num: The number of points to generate.
    """

    start: float
    stop: float
    step: int = 1

    def __post_init__(self) -> None:
        self.num: int = int((self.stop - self.start) / self.step)

    def _generate_samples(self, backend: AbstractBackend) -> Any:
        """Generates the linearly spaced points using arange."""
        return backend.arange(self.start, self.stop, self.step)


@dataclass
class ConstantSpace(SpaceType):
    """
    Represents a constant value repeated multiple times.

    This class generates a tensor containing the same constant value repeated
    `num` times.

    Attributes:
        value: The constant float value to be used.
        num: The number of times the constant value should be repeated. Defaults to 1.
    """

    value: float
    num: int | None = None

    def __post_init__(self) -> None:
        if self.num is None:
            object.__setattr__(self, "num", 1)

    def _generate_samples(self, backend: AbstractBackend):
        """Generates a tensor with the constant value repeated."""
        return backend.full((self.num,), self.value)


@dataclass
class ListSpace(SpaceType):
    """
    Represents a predefined list of values.

    This class generates a tensor directly from a provided list of float values.
    The number of points generated is equal to the length of the input list.

    Attributes:
        values: A list of float values to be converted into a tensor.
        num: The number of points, automatically set to length of values.
    """

    values: list[float]
    num: int = field(init=False)

    def __post_init__(self) -> None:
        object.__setattr__(self, "num", len(self.values))

    def _generate_samples(self, backend: AbstractBackend):
        """Generates a tensor from the provided list of values."""
        return backend.tensor(self.values)


@dataclass
class Distribution(SpaceType):
    """Base class for all distribution spaces."""

    ...


@dataclass
class NormalDistribution(Distribution):
    """
    Represents a sampling from a normal (Gaussian) distribution.

    This class generates a tensor of `num` random numbers sampled from a normal
    distribution with the specified `mean` and standard deviation `std`.

    Attributes:
        mean: The mean (center) of the normal distribution.
        std: The standard deviation (spread or width) of the normal distribution.
        num: The number of samples to generate.
    """

    mean: float
    std: float
    num: int

    def _generate_samples(self, backend: AbstractBackend):
        """Generates samples from the normal distribution."""
        return backend.normal(mean=self.mean, std=self.std, size=(self.num,))


@dataclass
class UniformDistribution(Distribution):
    """
    Represents a sampling from a uniform distribution.

    This class generates a tensor of `num` random numbers sampled from a uniform
    distribution over the interval [`low`, `high`).

    Attributes:
        low: The lower boundary of the output interval.
        high: The upper boundary of the output interval.
        num: The number of samples to generate.
    """

    low: float
    high: float
    num: int

    def _generate_samples(self, backend: AbstractBackend):
        """Generates samples from the uniform distribution."""
        # torch.rand samples from [0, 1), so we scale and shift.
        return backend.rand(size=(self.num,)) * (self.high - self.low) + self.low



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from einmesh._backends import NumpyBackend
from einmesh._einmesher import _EinMesher
from einmesh._parser import _einmesh
from einmesh.spaces import SpaceType

if TYPE_CHECKING:
    import numpy as np  # pyright: ignore[reportMissingImports]


def einmesh(
    pattern: str, *unamed_spaces: SpaceType, **named_spaces: SpaceType
) -> np.ndarray[Any, Any] | tuple[np.ndarray[Any, Any], ...]:
    """
    Numpy einmesh function.
    """
    return _einmesh(pattern, *unamed_spaces, backend=NumpyBackend(), **named_spaces)


# Expose a preconfigured EinMesher class bound to NumPy backend
class EinMesher(_EinMesher):
    """
    EinMesher bound to the NumPy backend. By default `.mesh()` will return NumPy arrays
    without needing to pass a backend.
    """

    def __init__(self, pattern: str, **spaces: SpaceType) -> None:
        super().__init__(pattern, backend=NumpyBackend(), **spaces)

    def sample(self) -> np.ndarray[Any, Any] | tuple[np.ndarray[Any, Any], ...]:
        return super().sample()



from typing import Any

from ._backends import AbstractBackend
from ._parser import KwargValueType, _einmesh


class _EinMesher:
    def __init__(
        self,
        pattern: str,
        *unamed_spaces: KwargValueType,
        backend: AbstractBackend | str = "numpy",
        **named_spaces: KwargValueType,
    ) -> None:
        self.pattern: str = pattern
        self.backend: AbstractBackend | str = backend
        self.named_spaces: dict[str, KwargValueType] = named_spaces
        self.unamed_spaces: tuple[KwargValueType, ...] = unamed_spaces

    def sample(self) -> Any | tuple[Any, ...]:
        return _einmesh(self.pattern, *self.unamed_spaces, backend=self.backend, **self.named_spaces)



from numpy.testing import assert_allclose

from einmesh import (
    ConstantSpace,
    LinSpace,
    ListSpace,
    NormalDistribution,
    UniformDistribution,
)
from einmesh._backends import AbstractBackend
from einmesh._einmesher import _EinMesher
from tests.conftest import parametrize_backends


@parametrize_backends
def test_einmesher_init(backend_cls: type[AbstractBackend]):
    """Test EinMesher initialization."""
    pattern = "x y"
    spaces = {"x": LinSpace(0, 1, 3), "y": ConstantSpace(5)}
    backend = backend_cls()
    mesher = _EinMesher(pattern, backend=backend, **spaces)
    assert mesher.pattern == pattern
    assert mesher.named_spaces == spaces
    assert mesher.unamed_spaces == ()


@parametrize_backends
def test_einmesher_basic_mesh(backend_cls: type[AbstractBackend]):
    """Test basic mesh generation (tuple output)."""
    backend = backend_cls()
    mesher = _EinMesher(
        "x y",
        x=LinSpace(0, 1, 3),
        y=ConstantSpace(10.0),
        backend=backend,
    )
    x_coords, y_coords = mesher.sample()

    expected_x = backend.tensor([[0.0], [0.5], [1.0]])
    expected_y = backend.tensor([[10.0], [10.0], [10.0]])

    assert backend.is_appropriate_type(x_coords)
    assert backend.is_appropriate_type(y_coords)
    assert_allclose(x_coords, expected_x)
    assert_allclose(y_coords, expected_y)
    assert backend.shape(x_coords) == (3, 1)
    assert backend.shape(y_coords) == (3, 1)


@parametrize_backends
def test_einmesher_stacked_mesh(backend_cls: type[AbstractBackend]):
    """Test stacked mesh generation (single tensor output)."""
    backend = backend_cls()
    mesher = _EinMesher(
        "* x y",
        x=LinSpace(0, 1, 3),
        y=ListSpace([10.0, 20.0]),
        backend=backend,
    )
    stacked_grid = mesher.sample()

    expected_grid = backend.tensor([
        [[0.0, 0.0], [0.5, 0.5], [1.0, 1.0]],  # x coords duplicated for y
        [[10.0, 20.0], [10.0, 20.0], [10.0, 20.0]],  # y coords broadcasted for x
    ])

    assert backend.is_appropriate_type(stacked_grid)
    assert backend.shape(stacked_grid) == (2, 3, 2)
    assert_allclose(stacked_grid, expected_grid)


@parametrize_backends
def test_einmesher_duplicate_names(backend_cls: type[AbstractBackend]):
    """Test mesh generation with duplicate names."""
    backend = backend_cls()
    mesher = _EinMesher("x x y", x=LinSpace(0, 1, 2), y=ConstantSpace(5), backend=backend)
    x0_coords, x1_coords, y_coords = mesher.sample()

    expected_x0 = backend.tensor([[[0.0], [0.0]], [[1.0], [1.0]]])  # Shape (2, 2, 1)
    expected_x1 = backend.tensor([[[0.0], [1.0]], [[0.0], [1.0]]])  # Shape (2, 2, 1)
    expected_y = backend.tensor([[[5.0], [5.0]], [[5.0], [5.0]]])  # Shape (2, 2, 1)

    assert backend.is_appropriate_type(x0_coords)
    assert backend.is_appropriate_type(x1_coords)
    assert backend.is_appropriate_type(y_coords)
    assert_allclose(x0_coords, expected_x0)
    assert_allclose(x1_coords, expected_x1)
    assert_allclose(y_coords, expected_y)
    assert backend.shape(x0_coords) == (2, 2, 1)
    assert backend.shape(x1_coords) == (2, 2, 1)
    assert backend.shape(y_coords) == (2, 2, 1)


@parametrize_backends
def test_einmesher_grouping(backend_cls: type[AbstractBackend]):
    """Test mesh generation with grouping."""
    backend = backend_cls()
    mesher = _EinMesher("x (y z)", x=LinSpace(0, 1, 2), y=LinSpace(10, 20, 3), z=ConstantSpace(5), backend=backend)
    x_coords, y_coords, z_coords = mesher.sample()

    expected_x = backend.tensor([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])
    expected_y = backend.tensor([[10.0, 15.0, 20.0], [10.0, 15.0, 20.0]])
    expected_z = backend.tensor([[5.0, 5.0, 5.0], [5.0, 5.0, 5.0]])

    assert backend.is_appropriate_type(x_coords)
    assert backend.is_appropriate_type(y_coords)
    assert backend.is_appropriate_type(z_coords)
    assert backend.shape(x_coords) == (2, 3)
    assert backend.shape(y_coords) == (2, 3)
    assert backend.shape(z_coords) == (2, 3)
    assert_allclose(x_coords, expected_x)
    assert_allclose(y_coords, expected_y)
    assert_allclose(z_coords, expected_z)


@parametrize_backends
def test_einmesher_stacked_grouping(backend_cls: type[AbstractBackend]):
    """Test stacked mesh generation with grouping."""
    backend = backend_cls()
    mesher = _EinMesher("* (x y)", x=LinSpace(0, 1, 2), y=ListSpace([10, 20]), backend=backend)
    stacked_grouped = mesher.sample()

    expected_stacked_grouped = backend.tensor([
        [0.0, 0.0, 1.0, 1.0],  # Flattened x_mesh
        [10.0, 20.0, 10.0, 20.0],  # Flattened y_mesh
    ])

    assert backend.is_appropriate_type(stacked_grouped)
    assert backend.shape(stacked_grouped) == (2, 4)
    assert_allclose(stacked_grouped, expected_stacked_grouped)


@parametrize_backends
def test_einmesher_random_resampling(backend_cls: type[AbstractBackend]):
    """Verify that random spaces are resampled on each mesh() call."""
    backend = backend_cls()
    mesher = _EinMesher(
        "norm uni const",
        norm=NormalDistribution(mean=0, std=1, num=5),
        uni=UniformDistribution(low=10, high=20, num=5),
        const=ConstantSpace(100),
        backend=backend,
    )

    # First call
    norm1, uni1, const1 = mesher.sample()

    # Second call
    norm2, uni2, const2 = mesher.sample()

    # Check shapes are consistent
    assert backend.shape(norm1) == backend.shape(norm2)
    assert backend.shape(uni1) == backend.shape(uni2)
    assert backend.shape(const1) == backend.shape(const2)

    # Check that random spaces generated different values
    assert backend.any(norm1 != norm2)
    assert backend.any(uni1 != uni2)

    # Check that constant space remained the same
    assert_allclose(const1, const2)


@parametrize_backends
def test_einmesher_stacked_random_resampling(backend_cls: type[AbstractBackend]):
    """Verify resampling for stacked random spaces."""
    backend = backend_cls()
    mesher = _EinMesher(
        "* norm uni",
        norm=NormalDistribution(mean=5, std=2, num=6),
        uni=UniformDistribution(low=-10, high=0, num=6),
        backend=backend,
    )

    # First call
    stacked1 = mesher.sample()

    # Second call
    stacked2 = mesher.sample()

    # Check shapes
    assert backend.shape(stacked1) == backend.shape(stacked2)
    assert backend.shape(stacked1)[0] == 2  # Should have 2 layers stacked

    # Check that the layers corresponding to random spaces are different
    assert backend.any(stacked1[0] != stacked2[0])  # norm layer
    assert backend.any(stacked1[1] != stacked2[1])  # uni layer


# Example of how to test potential errors (though most are in _einmesh)
# @parametrize_backends
# def test_einmesher_invalid_pattern(backend):
#     backend_instance = backend()
#     with pytest.raises(SomeExpectedError):
#         mesher = EinMesher("x -> y", x=LinSpace(0,1,2))
#         mesher.sample(backend=backend.framework_name)



from abc import abstractmethod

from einmesh._backends import AbstractBackend


class BackendOperator:
    """Base class for all operators acting on einmesh tensors."""

    @abstractmethod
    def _apply(self, x, backend: AbstractBackend): ...


class BackendAbs(BackendOperator):
    """Operator for absolute value."""

    def _apply(self, x, backend: AbstractBackend):
        return backend.abs(x)


class BackendAdd(BackendOperator):
    """Operator for addition."""

    def __init__(self, value: float):
        self.value = value

    def _apply(self, x, backend: AbstractBackend):
        return backend.add(x, summand=self.value)


class BackendSub(BackendOperator):
    """Operator for subtraction."""

    def __init__(self, value: float):
        self.value = value

    def _apply(self, x, backend: AbstractBackend):
        return backend.sub(x, subtrahend=self.value)


class BackendMul(BackendOperator):
    """Operator for multiplication."""

    def __init__(self, value: float):
        self.value = value

    def _apply(self, x, backend: AbstractBackend):
        return backend.mul(x, multiplier=self.value)


class BackendDiv(BackendOperator):
    """Operator for division."""

    def __init__(self, value: float):
        self.value = value

    def _apply(self, x, backend: AbstractBackend):
        return backend.truediv(x, divisor=self.value)


class BackendMod(BackendOperator):
    """Operator for modulo."""

    def __init__(self, value: float):
        self.value = value

    def _apply(self, x, backend: AbstractBackend):
        return backend.mod(x, divisor=self.value)


class BackendFloorDiv(BackendOperator):
    """Operator for floor division."""

    def __init__(self, value: float):
        self.value = value

    def _apply(self, x, backend: AbstractBackend):
        return backend.floordiv(x, divisor=self.value)


class BackendPow(BackendOperator):
    """Operator for power."""

    def __init__(self, exponent: float):
        self.exponent = exponent

    def _apply(self, x, backend: AbstractBackend):
        return backend.pow(base=x, exponent=self.exponent)


class BackendNeg(BackendOperator):
    """Operator for negation."""

    def _apply(self, x, backend: AbstractBackend):
        return backend.neg(x)


class BackendPos(BackendOperator):
    """Operator for positive value."""

    def _apply(self, x, backend: AbstractBackend):
        return backend.pos(x)


class BackendCos(BackendOperator):
    def _apply(self, x, backend: AbstractBackend):
        return backend.cos(x)


class BackendSin(BackendOperator):
    def _apply(self, x, backend: AbstractBackend):
        return backend.sin(x)


class BackendExp(BackendOperator):
    def _apply(self, x, backend: AbstractBackend):
        return backend.exp(x)


def cos(space):
    return space._with_operator(BackendCos())


def sin(space):
    return space._with_operator(BackendSin())


def exp(space):
    return space._with_operator(BackendExp())


class OperatorFactory:
    """Factory for creating operators."""

    @staticmethod
    def abs() -> BackendAbs:
        return BackendAbs()

    @staticmethod
    def add(value: float) -> BackendAdd:
        return BackendAdd(value=value)

    @staticmethod
    def sub(value: float) -> BackendSub:
        return BackendSub(value=value)

    @staticmethod
    def mul(value: float) -> BackendMul:
        return BackendMul(value=value)

    @staticmethod
    def div(value: float) -> BackendDiv:
        return BackendDiv(value=value)

    @staticmethod
    def mod(value: float) -> BackendMod:
        return BackendMod(value=value)

    @staticmethod
    def floor_div(value: float) -> BackendFloorDiv:
        return BackendFloorDiv(value=value)

    @staticmethod
    def pow(exponent: float) -> BackendPow:
        return BackendPow(exponent=exponent)

    @staticmethod
    def neg() -> BackendNeg:
        return BackendNeg()

    @staticmethod
    def pos() -> BackendPos:
        return BackendPos()



class PatternError(ValueError):
    """Error raised when the pattern is invalid."""

    def __init__(self, message: str) -> None:
        super().__init__(f"Invalid pattern: {message}")


class UnbalancedParenthesesError(PatternError):
    """Error raised when the pattern has unbalanced parentheses."""

    def __init__(self) -> None:
        super().__init__("Unbalanced parentheses in pattern")


class MultipleStarError(PatternError):
    """Error raised when multiple '*' are found in the output pattern."""

    def __init__(self) -> None:
        super().__init__("Multiple '*' are not allowed in the output pattern")


class UnderscoreError(PatternError):
    """Error raised when an underscore is found in the pattern."""

    def __init__(self) -> None:
        super().__init__("Underscores are not allowed in pattern names")


class ArrowError(PatternError):
    """Error raised when an arrow is found in the pattern."""

    def __init__(self) -> None:
        super().__init__("Arrow '->' is not allowed in pattern")


class MultipleEllipsisError(PatternError):
    """Error raised when multiple ellipsis are found in the pattern."""

    def __init__(self) -> None:
        super().__init__("Multiple ellipsis are not allowed in the einmesh pattern")


class UndefinedSpaceError(ValueError):
    """Error raised when a required sample space is not defined."""

    def __init__(self, space_name: str) -> None:
        super().__init__(f"Undefined space: {space_name}")


class InvalidListTypeError(TypeError):
    """Error raised when a list kwarg contains non-numeric types."""

    def __init__(self, space_name: str, invalid_types: list[str]) -> None:
        types_str = ", ".join(invalid_types)
        message = (
            f"List provided for space '{space_name}' must contain only int or float values, "
            f"but got types: [{types_str}]"
        )
        super().__init__(message)


class UnsupportedSpaceTypeError(TypeError):
    """Error raised when a kwarg value has an unsupported type."""

    def __init__(self, space_name: str, invalid_type: str) -> None:
        message = (
            f"Unsupported type for space '{space_name}': {invalid_type}. "
            f"Must be a SpaceType, int, float, or list[int | float]."
        )
        super().__init__(message)


class UnknownBackendError(Exception):
    """Error raised when a backend is unknown to einmesh."""

    def __init__(self, backend: str):
        self.message: str = f"Backend unknown to einmesh: {backend}"
        super().__init__(self.message)


class BackendNotInstalledError(Exception):
    """Error raised when a backend is not installed."""

    def __init__(self, backend: str):
        self.message: str = f"Backend {backend} not installed, but implicitly imported through einmesh"
        super().__init__(self.message)



import pytest

from einmesh._backends import AbstractBackend
from einmesh._exceptions import (
    InvalidListTypeError,
    UndefinedSpaceError,
    UnsupportedSpaceTypeError,
)
from einmesh._parser import _einmesh as einmesh
from einmesh.spaces import LinSpace
from tests.conftest import parametrize_backends


@parametrize_backends
def test_einmesh_basic(backend_cls: type[AbstractBackend]):
    """Test the basic functionality of einmesh without output pattern."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)

    meshes = einmesh("x y", x=x_space, y=y_space, backend=backend)

    assert len(meshes) == 2
    assert backend.is_appropriate_type(meshes[0])
    assert backend.is_appropriate_type(meshes[1])
    assert backend.shape(meshes[0]) == (5, 3)
    assert backend.shape(meshes[1]) == (5, 3)


@parametrize_backends
def test_einmesh_star_pattern(backend_cls: type[AbstractBackend]):
    """Test einmesh with * pattern to stack all meshes."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)
    z_space = LinSpace(0.0, 1.0, 2)

    # Using just *
    result = einmesh("* x y z", x=x_space, y=y_space, z=z_space, backend=backend)

    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (3, 5, 3, 2)  # 3 meshes stacked as first dimension

    # Check that the result contains the original meshes
    x_mesh, y_mesh, z_mesh = einmesh("x y z", x=x_space, y=y_space, z=z_space, backend=backend)
    assert backend.allclose(result[0], x_mesh)
    assert backend.allclose(result[1], y_mesh)
    assert backend.allclose(result[2], z_mesh)


@parametrize_backends
def test_einmesh_parentheses_pattern(backend_cls: type[AbstractBackend]):
    """Test einmesh with parentheses pattern to reshape dimensions."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)

    # Using pattern with parentheses
    result = einmesh("(x y)", x=x_space, y=y_space, backend=backend)

    # Basic check that we get a tensor
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert backend.is_appropriate_type(result[0])
    assert backend.is_appropriate_type(result[1])

    # Using einops.rearrange properly flattens the dimensions within parentheses
    # So the result should be a 1D tensor with shape (5*3,) = (15,)
    assert backend.shape(result[0]) == (5 * 3,)
    assert backend.shape(result[1]) == (5 * 3,)


@parametrize_backends
def test_einmesh_output_dimension_ordering(backend_cls: type[AbstractBackend]):
    """Test that einmesh respects dimension ordering in output pattern."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)

    # Get original meshes
    x1_mesh, y1_mesh = einmesh("x y", x=x_space, y=y_space, backend=backend)
    y2_mesh, x2_mesh = einmesh("y x", x=x_space, y=y_space, backend=backend)

    # Ensure results are transposed
    assert backend.allclose(x1_mesh, x2_mesh.transpose(1, 0))
    assert backend.allclose(y1_mesh, y2_mesh.transpose(1, 0))


@parametrize_backends
def test_star_position(backend_cls: type[AbstractBackend]):
    """Test that einmesh handles star position correctly."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 7)
    y_space = LinSpace(0.0, 1.0, 9)

    result = einmesh("* x y", x=x_space, y=y_space, backend=backend)
    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (2, 7, 9)

    result = einmesh("x * y", x=x_space, y=y_space, backend=backend)
    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (7, 2, 9)

    result = einmesh("x y *", x=x_space, y=y_space, backend=backend)
    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (7, 9, 2)


@parametrize_backends
def test_axis_collection(backend_cls: type[AbstractBackend]):
    """Test that einmesh handles axis collection correctly."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)

    result = einmesh("* (x y)", x=x_space, y=y_space, backend=backend)
    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (2, 5 * 3)

    result = einmesh("(x y) *", x=x_space, y=y_space, backend=backend)
    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (5 * 3, 2)


@parametrize_backends
def test_star_in_axis_collection(backend_cls: type[AbstractBackend]):
    """Test that einmesh handles star in axis collection correctly."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)

    result = einmesh("(* x) y", x=x_space, y=y_space, backend=backend)
    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (2 * 5, 3)


@parametrize_backends
def test_invalid_pattern(backend_cls: type[AbstractBackend]):
    """Test that einmesh raises error for invalid patterns."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)

    with pytest.raises(UndefinedSpaceError):
        einmesh("x y", x=x_space, backend=backend)  # Missing y space


@parametrize_backends
def test_einmesh_auto_conversion(backend_cls: type[AbstractBackend]):
    """Test automatic conversion of int, float, and list kwargs."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 3)

    # Test int -> ConstantSpace
    x_coords, y_coords = einmesh("x y", x=x_space, y=5, backend=backend)
    assert backend.is_appropriate_type(y_coords)
    assert backend.shape(y_coords) == (3, 1)
    assert backend.all(y_coords == 5.0)

    # Test float -> ConstantSpace
    x_coords, z_coords = einmesh("x z", x=x_space, z=-2.5, backend=backend)
    assert backend.is_appropriate_type(z_coords)
    assert backend.shape(z_coords) == (3, 1)
    assert backend.all(z_coords == -2.5)

    # Test list[int] -> ListSpace
    list_int = [1, 2, 3, 4]
    x_coords, w_coords = einmesh("x w", x=x_space, w=list_int, backend=backend)  # type: ignore[arg-type]
    assert backend.is_appropriate_type(w_coords)
    assert backend.shape(w_coords) == (3, 4)
    assert backend.allclose(w_coords[0], backend.tensor(list_int, dtype=backend.float))

    # Test list[float] -> ListSpace
    list_float = [1.1, 2.2, 3.3]
    x_coords, v_coords = einmesh("x v", x=x_space, v=list_float, backend=backend)
    assert backend.is_appropriate_type(v_coords)
    assert backend.shape(v_coords) == (3, 3)
    assert backend.allclose(v_coords[0], backend.tensor(list_float, dtype=backend.float))

    # Test mixed list[int | float] -> ListSpace
    list_mixed = [1, 2.5, 3]
    x_coords, u_coords = einmesh("x u", x=x_space, u=list_mixed, backend=backend)
    assert backend.is_appropriate_type(u_coords)
    assert backend.shape(u_coords) == (3, 3)
    assert backend.allclose(u_coords[0], backend.tensor(list_mixed, dtype=backend.float))

    # Test combination with explicit SpaceType and stacking
    stacked_res = einmesh("* x y z", x=x_space, y=100, z=[10, 20], backend=backend)
    assert backend.is_appropriate_type(stacked_res)
    assert backend.shape(stacked_res) == (3, 3, 1, 2)  # Stack, x, y (const), z (list)
    assert backend.all(stacked_res[1] == 100.0)
    assert backend.allclose(stacked_res[2, 0, 0, :], backend.tensor([10.0, 20.0]))


@parametrize_backends
def test_einmesh_auto_conversion_errors(backend_cls: type[AbstractBackend]):
    """Test errors raised during automatic type conversion."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 3)

    # Test invalid type (string)
    with pytest.raises(UnsupportedSpaceTypeError) as exc_info_type:
        einmesh("x y", x=x_space, y="not a number", backend=backend)  # type: ignore[arg-type] # Intentionally passing invalid type
    assert "Unsupported type for space 'y': str" in str(exc_info_type.value)

    # Test list with invalid contents (string)
    with pytest.raises(InvalidListTypeError) as exc_info_list:
        einmesh("x z", x=x_space, z=[1, 2, "three"], backend=backend)  # type: ignore[arg-type, list-item] # Intentionally passing invalid list item type
    assert "List provided for space 'z' must contain only int or float" in str(exc_info_list.value)
    # Check that the message correctly identifies the invalid type ('str')
    assert "got types: [str]" in str(exc_info_list.value)

    x_coords_empty, w_coords_empty = einmesh("x w", x=x_space, w=[], backend=backend)
    assert backend.shape(w_coords_empty) == (3, 0)  # Shape should reflect the empty list dimension


@parametrize_backends
def test_ellipsis_basic_substitution(backend_cls: type[AbstractBackend]):
    """Test basic ellipsis substitution without stacking or grouping."""
    backend = backend_cls()

    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)
    args = [x_space, y_space]

    x_mesh, y_mesh = einmesh("...", *args, backend=backend)
    assert backend.is_appropriate_type(x_mesh)
    assert backend.is_appropriate_type(y_mesh)
    assert backend.shape(x_mesh) == (5, 3)
    assert backend.shape(y_mesh) == (5, 3)


@parametrize_backends
def test_ellipsis_stacked_substitution(backend_cls: type[AbstractBackend]):
    """Test ellipsis substitution with stacking."""
    backend = backend_cls()

    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)
    args = [x_space, y_space]

    result = einmesh("* ...", *args, backend=backend)
    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (2, 5, 3)


@parametrize_backends
def test_ellipsis_grouped_substitution(backend_cls: type[AbstractBackend]):
    """Test ellipsis substitution with stacking and grouping."""
    backend = backend_cls()

    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)
    args = [x_space, y_space]

    result = einmesh("* (...)", *args, backend=backend)
    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (2, 15)


@parametrize_backends
def test_ellipsis_combined_axes(backend_cls: type[AbstractBackend]):
    """Test ellipsis substitution combined with other axes."""
    backend = backend_cls()

    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)
    z_space = LinSpace(0.0, 1.0, 4)
    args = [x_space, y_space, z_space]

    result = einmesh("* x y ... z", *args, x=x_space, y=y_space, z=z_space, backend=backend)
    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (6, 5, 3, 5, 3, 4, 4)


@parametrize_backends
def test_ellipsis_combined_axes_with_parentheses(backend_cls: type[AbstractBackend]):
    """Test ellipsis substitution combined with other axes."""
    backend = backend_cls()

    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)
    z_space = LinSpace(0.0, 1.0, 4)
    args = [x_space, y_space, z_space]

    result = einmesh("* x y (...) z", *args, x=x_space, y=y_space, z=z_space, backend=backend)
    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (6, 5, 3, 5 * 3 * 4, 4)



import re
from typing import Any

import einops

from einmesh._backends import AbstractBackend, JaxBackend, NumpyBackend, TorchBackend
from einmesh._exceptions import (
    ArrowError,
    InvalidListTypeError,
    MultipleEllipsisError,
    MultipleStarError,
    UnbalancedParenthesesError,
    UndefinedSpaceError,
    UnderscoreError,
    UnknownBackendError,
    UnsupportedSpaceTypeError,
)
from einmesh.spaces import ConstantSpace, ListSpace, SpaceType


def _handle_duplicate_names(
    pattern: str,
    shape_pattern: str,
    kwargs: dict[str, SpaceType],
) -> tuple[str, dict[str, str], dict[str, SpaceType]]:
    """
    Handles renaming of duplicate space names in the pattern and updates kwargs.

    If a space name appears multiple times in the pattern (e.g., "x x y"),
    it renames subsequent occurrences with an index suffix (e.g., "x_0 x_1 y").
    It updates the pattern string and the kwargs dictionary accordingly, adding
    entries for the new names and removing the original duplicate entries from kwargs.

    Args:
        pattern: The original einmesh pattern string.
        shape_pattern: The pattern string with parentheses removed.
        kwargs: The dictionary of space names to SpaceType objects.

    Returns:
        A tuple containing:
            - The modified pattern string with duplicates renamed.
            - A dictionary mapping new names to original names (e.g., {"x_0": "x", "x_1": "x"}).
            - The updated kwargs dictionary with renamed keys and removed originals.
    """
    seen_names: dict[str, int] = {}
    name_mapping: dict[str, str] = {}

    # First count occurrences of each name (excluding '*')
    for name in shape_pattern.split():
        if name != "*":
            seen_names[name] = seen_names.get(name, 0) + 1

    # Then rename duplicates for each unique name with counts > 1
    for name in list(seen_names.keys()):
        if seen_names[name] > 1:
            for i in range(seen_names[name]):
                new_name = f"{name}_{i}"
                # Use regex to replace only whole words to avoid partial matches
                pattern = re.sub(rf"\b{name}\b", new_name, pattern, count=1)
                name_mapping[new_name] = name

    # Update kwargs with renamed spaces
    updated_kwargs = kwargs.copy()  # Avoid modifying the original dict directly
    for new_name, orig_name in name_mapping.items():
        if orig_name in updated_kwargs:
            updated_kwargs[new_name] = updated_kwargs[orig_name]

    # Remove original names that were renamed
    orig_names_to_remove = set(name_mapping.values())
    for orig_name in orig_names_to_remove:
        if orig_name in updated_kwargs:
            updated_kwargs.pop(orig_name)

    return pattern, name_mapping, updated_kwargs


# Define the type alias for acceptable kwarg values
KwargValueType = SpaceType | int | float | list[int | float]


def _get_backend(backend: AbstractBackend | str) -> AbstractBackend:
    if isinstance(backend, str):
        if backend == "torch":
            return TorchBackend()
        elif backend == "jax":
            return JaxBackend()
        elif backend == "numpy":
            return NumpyBackend()
        else:
            raise UnknownBackendError(backend)
    elif isinstance(backend, AbstractBackend):
        return backend
    else:
        raise UnknownBackendError(backend)


def _parse_args(args: tuple[KwargValueType, ...]) -> tuple[str, dict[str, KwargValueType]]:
    """
    Parses the arguments and returns a tuple of processed args and kwargs.
    """
    processed_args: dict[str, KwargValueType] = {}
    for i, arg in enumerate(args):
        args_name = f"arg__{i}"
        processed_args[args_name] = arg

    elipsis_substituion = " ".join(list(processed_args.keys()))

    return elipsis_substituion, processed_args


def _parse_kwargs(kwargs: dict[str, KwargValueType]) -> dict[str, SpaceType]:
    """
    Parses the kwargs and returns a dictionary of SpaceType objects.
    """
    processed_kwargs: dict[str, SpaceType] = {}
    for name, value in kwargs.items():
        if isinstance(value, (int, float)):
            processed_kwargs[name] = ConstantSpace(value=float(value))
        elif isinstance(value, list):
            # Check if list contains only numbers (int or float)
            if all(isinstance(item, (int, float)) for item in value):
                # Convert all items to float for consistency
                processed_kwargs[name] = ListSpace(values=[float(item) for item in value])
            else:
                # Explicitly annotate the list as list[str] to satisfy the type checker
                invalid_types: list[str] = [type(item).__name__ for item in value if not isinstance(item, (int, float))]
                raise InvalidListTypeError(space_name=name, invalid_types=invalid_types)
        elif isinstance(value, SpaceType):
            processed_kwargs[name] = value
        else:
            raise UnsupportedSpaceTypeError(space_name=name, invalid_type=type(value).__name__)
    return processed_kwargs


def _einmesh(
    pattern: str, *args: KwargValueType, backend: AbstractBackend | str = "numpy", **kwargs: KwargValueType
) -> Any | tuple[Any, ...]:
    """
    Creates multi-dimensional meshgrids using an einops-style pattern string.

    `einmesh` simplifies the creation and manipulation of multi-dimensional
    meshgrids by specifying sampling spaces and their arrangement using an
    intuitive pattern inspired by `einops`.

    The pattern string defines the dimensions and structure of the output:
    - **Space Names:** Correspond to keyword arguments providing sampling definitions.
    - **Sampling Definitions:** Can be:
        - `SpaceType` objects (e.g., `LinSpace`, `LogSpace`).
        - Single `int` or `float` values (automatically converted to `ConstantSpace`).
        - `list` of `int` or `float` values (automatically converted to `ListSpace`).
      (e.g., `x=LinSpace(0, 1, 5)`, `y=10.0`, `z=[1, 2, 4]`)
    - **Repeated Names:** Handles dimensions derived from the same space type
      (e.g., "x x y" results in dimensions named `x_0`, `x_1`, `y`).
    - **Stacking (`*`):** Stacks the generated meshgrids for each space along a new
      dimension. Only one `*` is allowed. The output is a single tensor.
    - **Grouping (`()`):** Groups dimensions together in the output tensor shape,
      affecting the `einops.rearrange` operation applied internally.

    If the pattern does *not* contain `*`, the function returns a tuple of tensors,
    one for each space name in the pattern (after handling duplicates). Each tensor
    in the tuple represents the coordinates for that specific dimension across the
    entire meshgrid, ordered according to the pattern.

    If the pattern *does* contain `*`, the function returns a single tensor where
    the individual meshgrids are stacked along the dimension specified by `*`.

    Examples:
        >>> from einmesh import LinSpace, einmesh
        >>> x_space = LinSpace(0, 1, 5)

        >>> # Basic 2D meshgrid (tuple output)
        >>> x_coords, y_coords = einmesh("x y", x=x_space, y=10.0)
        >>> x_coords.shape
        torch.Size([5, 1])
        >>> y_coords.shape
        torch.Size([5, 1])
        >>> y_coords # Constant value 10.0 repeated
        tensor([[10.],
                [10.],
                [10.],
                [10.],
                [10.]])

        >>> # Using a list for a dimension
        >>> x_coords, z_coords = einmesh("x z", x=x_space, z=[1, 2, 4])
        >>> z_coords.shape
        torch.Size([5, 3])
        >>> z_coords[0] # First row shows the list values
        tensor([1., 2., 4.])

        >>> # Stacked meshgrid (single tensor output)
        >>> stacked_grid = einmesh("* x y", x=x_space, y=10.0)
        >>> stacked_grid.shape
        torch.Size([2, 5, 1])

        >>> # Grouping affects rearrangement
        >>> stacked_grouped = einmesh("* (x y)", x=x_space, y=[10.0, 20.0])
        >>> stacked_grouped.shape
        torch.Size([2, 10])

        >>> # Repeated spaces (using implicit ConstantSpace)
        >>> x0_coords, x1_coords = einmesh("x x", x=5)
        >>> x0_coords.shape
        torch.Size([1, 1])
        >>> x0_coords.item()
        5.0

    Args:
        pattern: The einops-style string defining meshgrid structure.
        **kwargs: Keyword arguments mapping space names in the pattern to
                  sampling definitions (`SpaceType`, `int`, `float`, or `list`).

    Returns:
        Union[torch.Tensor, tuple[torch.Tensor, ...]]: A `torch.Tensor` if the pattern includes `*` (stacking), or a
        `tuple[torch.Tensor, ...]` if the pattern does not include `*`.

    Raises:
        UnbalancedParenthesesError: If parentheses in the pattern are not balanced.
        MultipleStarError: If the pattern contains more than one `*`.
        UndefinedSpaceError: If a name in the pattern doesn't have a corresponding
                             kwarg definition.
        ArrowError: If the pattern contains '->', which is not supported.
        TypeError: If a kwarg value is not a `SpaceType`, `int`, `float`, or a
                   valid `list` of `int`/`float`.
    """

    _verify_pattern(pattern)
    backend: AbstractBackend = _get_backend(backend)

    elipsis_substituion, processed_args = _parse_args(args)

    pattern: str = pattern.replace("...", elipsis_substituion)

    kwargs: dict[str, KwargValueType] = {**kwargs, **processed_args}

    # Process kwargs to convert raw values to SpaceTypes
    processed_kwargs: dict[str, SpaceType] = _parse_kwargs(kwargs)

    # get stack index
    shape_pattern: str = pattern.replace("(", "").replace(")", "")
    stack_idx: int | None = shape_pattern.split().index("*") if "*" in shape_pattern else None

    # Check for and handle duplicate names in pattern using processed kwargs
    pattern, name_mapping, processed_kwargs_renamed = _handle_duplicate_names(pattern, shape_pattern, processed_kwargs)

    # Determine the final order of space names from the potentially modified pattern
    final_pattern_names: list[str] = pattern.replace("(", "").replace(")", "").split()
    # Filter out '*' as it's not a sampling space name
    sampling_list: list[str] = [name for name in final_pattern_names if name != "*"]

    # Pass the ordered sampling_list and processed+renamed kwargs
    meshes, dim_shapes = _generate_samples(sampling_list, backend, **processed_kwargs_renamed)

    # Handle star pattern for stacking meshes
    input_sampling_list: list[str] = list(sampling_list)  # Base list for input pattern
    if stack_idx is not None:
        meshes: Any = backend.stack(meshes, dim=stack_idx)
        dim_shapes["einstack"] = meshes.shape[stack_idx]
        # Insert 'einstack' into the sampling list copy at the correct index for the input pattern
        input_sampling_list.insert(stack_idx, "einstack")

    # Define the input pattern based on the actual order of dimensions in the tensor(s)
    input_pattern: str = " ".join(input_sampling_list)

    if backend.is_appropriate_type(meshes):  # Stacked case
        # Output pattern: User pattern with '*' replaced by 'einstack'
        output_pattern: str = pattern.replace("*", "einstack")
        meshes: Any = einops.rearrange(meshes, f"{input_pattern} -> {output_pattern}", **dim_shapes)

    elif isinstance(meshes, list):  # Non-stacked case (must be tuple eventually)
        rearranged_meshes = []
        # Output pattern: User pattern (with renames, no '*' or 'einstack')
        output_pattern = pattern
        # Input pattern is the same for all meshes in the list
        for mesh in meshes:
            # Rearrange each mesh individually
            rearranged_mesh = einops.rearrange(mesh, f"{input_pattern} -> {output_pattern}", **dim_shapes)
            rearranged_meshes.append(rearranged_mesh)
        meshes = tuple(rearranged_meshes)  # Convert list back to tuple

    return meshes


def _generate_samples(sampling_list: list[str], backend: AbstractBackend, **kwargs: SpaceType):
    """
    Generates 1D samples for each space and creates initial meshgrids.

    Uses `torch.meshgrid` with `indexing="ij"` based on the ordered
    `sampling_list`.

    Args:
        sampling_list: An ordered list of space names (potentially renamed
                       if duplicates existed in the original pattern).
        **kwargs: The dictionary of space names to `SpaceType` objects,
                  potentially updated with renamed keys.

    Returns:
        A tuple containing:
            - A list of `torch.Tensor` objects representing the meshgrid
              coordinates for each dimension in `sampling_list`. The order
              matches `sampling_list`.
            - A dictionary mapping space names to their corresponding dimension sizes.

    Raises:
        UndefinedSpaceError: If a name in `sampling_list` is not found in `kwargs`.
    """
    lin_samples: list[Any] = []
    dim_shapes: dict[str, int] = {}
    # Iterate using the provided sampling_list to ensure correct order
    for p in sampling_list:
        if p not in kwargs:
            # This check might be redundant if pattern validation is robust, but safer to keep
            raise UndefinedSpaceError(p)
        samples = kwargs[p]._sample(backend)
        lin_samples.append(samples)
        dim_shapes[p] = backend.shape(samples)[0]
    # The order of meshes returned by torch.meshgrid(indexing='ij')
    # corresponds to the order of tensors in lin_samples.
    meshes = list(backend.meshgrid(*lin_samples, indexing="ij"))
    return meshes, dim_shapes


def _verify_pattern(pattern: str) -> None:
    if pattern.count("*") > 1:
        raise MultipleStarError()
    if pattern.count("...") > 1:
        raise MultipleEllipsisError()
    if pattern.count("(") != pattern.count(")"):
        raise UnbalancedParenthesesError()
    if "_" in pattern:
        raise UnderscoreError()
    if "->" in pattern:
        raise ArrowError()



import pytest

from einmesh._backends import AbstractBackend, JaxBackend, NumpyBackend, TorchBackend, get_backend
from einmesh._einmesher import _EinMesher
from einmesh._exceptions import (
    ArrowError,
    MultipleStarError,
    UnbalancedParenthesesError,
    UnderscoreError,
    UnknownBackendError,
)
from einmesh.spaces import LinSpace
from tests.conftest import parametrize_backends

"""
Error condition tests for parser validation and backend selection.
"""

# Parser validation errors


@parametrize_backends
def test_multiple_star_error(backend_cls: type[AbstractBackend]):
    backend = backend_cls()
    mesher = _EinMesher(
        "x * y * z", x=LinSpace(0.0, 1.0, 2), y=LinSpace(0.0, 1.0, 2), z=LinSpace(0.0, 1.0, 2), backend=backend
    )
    with pytest.raises(MultipleStarError):
        mesher.sample()


@parametrize_backends
def test_unbalanced_parentheses_error(backend_cls: type[AbstractBackend]):
    backend = backend_cls()
    # Missing closing parenthesis
    mesher1 = _EinMesher("x (y z", x=LinSpace(0.0, 1.0, 2), y=LinSpace(0.0, 1.0, 2), backend=backend)
    with pytest.raises(UnbalancedParenthesesError):
        mesher1.sample()
    # Extra closing parenthesis
    mesher2 = _EinMesher("x y)", x=LinSpace(0.0, 1.0, 2), y=LinSpace(0.0, 1.0, 2), backend=backend)
    with pytest.raises(UnbalancedParenthesesError):
        mesher2.sample()


@parametrize_backends
def test_arrow_error(backend_cls: type[AbstractBackend]):
    backend = backend_cls()
    mesher = _EinMesher("x -> y", x=LinSpace(0.0, 1.0, 2), y=LinSpace(0.0, 1.0, 2), backend=backend)
    with pytest.raises(ArrowError):
        mesher.sample()


@parametrize_backends
def test_underscore_error(backend_cls: type[AbstractBackend]):
    backend = backend_cls()
    mesher = _EinMesher("x_y y", x=LinSpace(0.0, 1.0, 2), y=LinSpace(0.0, 1.0, 2), backend=backend)
    with pytest.raises(UnderscoreError):
        mesher.sample()


def test_unknown_backend_error():
    mesher = _EinMesher("x y", x=LinSpace(0.0, 1.0, 2), y=LinSpace(0.0, 1.0, 2), backend="NotDefinedBackend")
    with pytest.raises(UnknownBackendError):
        mesher.sample()


@pytest.mark.skipif(not TorchBackend.is_available(), reason="Torch not available")
def test_get_backend_selection_torch():
    import torch

    tensor = torch.tensor([1, 2, 3])
    backend = get_backend(tensor)
    assert isinstance(backend, TorchBackend)


@pytest.mark.skipif(not NumpyBackend.is_available(), reason="Numpy not available")
def test_get_backend_selection_numpy():
    import numpy as np

    array = np.array([1, 2, 3])
    backend = get_backend(array)
    assert isinstance(backend, NumpyBackend)


@pytest.mark.skipif(not JaxBackend.is_available(), reason="JAX not available")
def test_get_backend_selection_jax():
    import jax.numpy as jnp

    array = jnp.zeros((3,))
    backend = get_backend(array)
    assert isinstance(backend, JaxBackend)


# JAX backend seed initialization sanity check


@pytest.mark.skipif(not JaxBackend.is_available(), reason="JAX not available")
def test_jax_backend_seed_variation():
    # The initial PRNGKey should differ across instances
    b1 = JaxBackend()
    b2 = JaxBackend()
    # Convert PRNGKey arrays to Python lists for comparison
    key1 = b1.key.tolist()
    key2 = b2.key.tolist()
    assert key1 != key2



from ._operators import cos, exp, sin
from .spaces import (
    ConstantSpace,
    LinSpace,
    ListSpace,
    LogSpace,
    NormalDistribution,
    RangeSpace,
    SpaceType,
    UniformDistribution,
)

__all__ = [
    "ConstantSpace",
    "LinSpace",
    "ListSpace",
    "LogSpace",
    "NormalDistribution",
    "RangeSpace",
    "SpaceType",
    "UniformDistribution",
    "cos",
    "exp",
    "sin",
]



from typing import TYPE_CHECKING, Any, TypeAlias

from numpy.typing import NDArray

Number: TypeAlias = int | float | bool

if TYPE_CHECKING:
    import jax  # pyright: ignore[reportMissingImports]
    import torch  # pyright: ignore[reportMissingImports]

    Tensor: TypeAlias = NDArray[Any] | torch.Tensor | jax.Array
else:
    Tensor = Any



from einmesh._backends import AbstractBackend
from einmesh._parser import _einmesh as einmesh
from einmesh.spaces import LinSpace
from tests.conftest import parametrize_backends


@parametrize_backends
def test_repeated_space(backend_cls: type[AbstractBackend]):
    """Test that einmesh handles repeated spaces correctly."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    result = einmesh("* x x", x=x_space, backend=backend)
    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (2, 5, 5)


@parametrize_backends
def test_repeated_space_in_parentheses(backend_cls: type[AbstractBackend]):
    """Test that einmesh handles repeated spaces in parentheses correctly."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    result = einmesh("(x x)", x=x_space, backend=backend)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert backend.is_appropriate_type(result[0])
    assert backend.is_appropriate_type(result[1])
    assert backend.shape(result[0]) == (25,)
    assert backend.shape(result[1]) == (25,)


@parametrize_backends
def test_repeated_space_in_parentheses_and_star(backend_cls: type[AbstractBackend]):
    """Test that einmesh handles repeated spaces in parentheses and star correctly."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    result = einmesh("* (x x)", x=x_space, backend=backend)
    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (2, 25)


@parametrize_backends
def test_repeated_space_with_non_repeated_space(backend_cls: type[AbstractBackend]):
    """Test that einmesh handles repeated spaces with non-repeated spaces correctly."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)
    z_space = LinSpace(0.0, 1.0, 2)
    result = einmesh("* x (y y) z", x=x_space, y=y_space, z=z_space, backend=backend)
    assert backend.is_appropriate_type(result)
    assert backend.shape(result) == (4, 5, 3 * 3, 2)


@parametrize_backends
def test_very_repeated_space(backend_cls: type[AbstractBackend]):
    """Test that einmesh handles very repeated spaces correctly."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)
    results = einmesh("x x (x x) x (y x) x x", x=x_space, y=y_space, backend=backend)
    assert isinstance(results, tuple)
    assert len(results) == 9
    for result in results:
        assert backend.is_appropriate_type(result)
        assert backend.shape(result) == (5, 5, 5 * 5, 5, 3 * 5, 5, 5)


@parametrize_backends
def test_multi_repeated_space(backend_cls: type[AbstractBackend]):
    """Test that einmesh handles very repeated spaces with star correctly."""
    backend = backend_cls()
    x_space = LinSpace(0.0, 1.0, 5)
    y_space = LinSpace(0.0, 1.0, 3)
    z_space = LinSpace(0.0, 1.0, 2)
    results = einmesh("* x y z x y z x y z", x=x_space, y=y_space, z=z_space, backend=backend)
    assert backend.is_appropriate_type(results)
    assert backend.shape(results) == (9, 5, 3, 2, 5, 3, 2, 5, 3, 2)
